import matplotlib.pyplot as plt
import scienceplots

from tsfm.data import generator
from tsfm.models.armodel import ARModel
from tsfm.models.base import Model
from tsfm.models.chronos2 import Chronos2
from tsfm.models.moirai import Moirai
from tsfm.models.moirai2 import Moirai2

plt.style.use(["science", "no-latex", "notebook"])

__all__ = ["ARModel", "Chronos2", "Model", "Moirai", "Moirai2", "generator"]
